<?php
//
// +----------------------------------------------------------------------+
// | PHP version 4.0                                                      |
// +----------------------------------------------------------------------+
// | Copyright (c) 1997, 1998, 1999, 2000 The PHP Group                   |
// +----------------------------------------------------------------------+
// | This source file is subject to version 2.02 of the PHP license,      |
// | that is bundled with this package in the file LICENSE, and is        |
// | available at through the world-wide-web at                           |
// | http://www.php.net/license/2_02.txt.                                 |
// | If you did not receive a copy of the PHP license and are unable to   |
// | obtain it through the world-wide-web, please send a note to          |
// | license@php.net so we can mail you a copy immediately.               |
// +----------------------------------------------------------------------+
// | Authors: Stig Bakken <ssb@fast.no>                                   |
// |                                                                      |
// +----------------------------------------------------------------------+
//
// Database independent query interface definition for PHP's Virt
// extension.
//

//
// XXX legend:
//
// XXX ERRORMSG: The error message from the virt function should
//				 be registered here.
//

include_once('DB.php');
include_once('DB/common.php');

define("VIRT_TERMTYPE",			'T');
define("VIRT_TERMTOKEN",		';');

define("VIRT_ASSOC",			MYSQL_ASSOC);
define("VIRT_NUM",				MYSQL_NUM);
define("VIRT_BOTH",				MYSQL_BOTH);

define("VIRT_CHAR",				0);
define("VIRT_NUMERIC",			1);

define("VIRT_MAXLINE",			4096);

class DB_virt extends DB_common {
    // {{{ properties

//	var $connection;
	var $phptype, $dbsyntax;
	var $prepare_tokens = array();
	var $prepare_types = array();

	var $lexArr = array();

	var $sqlAction = '';

	// Per-table array with tableName as key
	var $tableFields = array();		// contain arrays of fields
	var $tableTypes = array();		// contain arrays of types indexed by field name
	var $tableAutoInc = array();	// contain auto-increase field names
	var $tableLastIns = array();	// contain last inserted value for the AutoInc fields
	var $tableRecords = array();	// contain arrays of record, which is array of field => value

	// Parse result (transient)
	var $tableName = '';
	var $fieldList = array();
	var $valueList = array();
	var $whereFuncBody = '';
	var $orderByList = array();  // array of (fld, desc, type)
	var $distinct = false;

	var $lastInsertID = 0;
	var $virtErrorNum = DB_OK;

	// Key is result identifier; Element is an array with the 'tableName' element,
	// the 'fieldList' element, which is an array of names of selected fields, and
	// the 'resultRecords' element, which is an array of keys to tableRecords[tablename]
	// The first two elements are null so a valid result starts with 2
	// to avoid freeResult freeing a true/false result.
	var $resultArray = array(NULL, NULL);

	var $loadedFileName = '';

	var $debugLevel = 0;

    // }}}

    // {{{ constructor

	/**
	 * DB_virt constructor.
	 *
	 * @access public
	 */
	function DB_virt() {
		$this->phptype = 'virt';
		$this->dbsyntax = 'virt';
		$this->features = array(
			'prepare' => false,
			'pconnect' => true,
			'transactions' => false
		);
	}

		// }}}


	function errorCode($nativecode)
	{
		return $nativecode;
	}


/************************************************************
 *
 * SQL processing functions
 *
 */

	function lexical($statmArg)
	{
		// A brut force lexical analyser

		// Preprocess the statement
		$statm = $statmArg;
		$statm = ereg_replace('\\\\,', '!Impossible!EscapedComma!', $statm);
		$statm = ereg_replace(',', ' ,', $statm);
		$statm = ereg_replace(VIRT_TERMTOKEN . '$', ' ' . VIRT_TERMTOKEN, $statm);
		$statm = ereg_replace('!Impossible!EscapedComma!', '\\,', $statm);
if ($this->debugLevel) printf("<b>function</b> %s::lexical statm=[$statm]<br>\n", get_class($this));
		// No double spaces
		do
		{
			$statmSaved = $statm;
			$statm = ereg_replace('  ', ' ', $statm);
		} while ($statm != $statmSaved);
//if ($this->debugLevel) printf("<b>function</b> %s::lexical statm=[$statm]<br>\n", get_class($this));

		$tokenArr = split(' ', $statm);
		$this->lexArr = array();
		$litSeg = '';  // Literal segment that was broken by space

		foreach($tokenArr as $token)
		{
			$lexType = '';
			$tokenTaken = $token;

			if ($litSeg)
			{
				$tokenTaken = $litSeg . ' ' . $token;
				$lexType = 'L';
//if ($this->debugLevel) printf("<b>function</b> %s::lexical litSeg tokenTaken=[$tokenTaken]<br>\n", get_class($this));
			}
			$litSeg = '';

			//$gotComma = false;
			if (empty($lexType))
			{
				// Separate the trailing comma (as it's not in a quoted string)
				//$tokenTaken = ereg_replace(',$', '', $token);
				//$gotComma = ($tokenTaken != $token);

				$lowerToken = strtolower($tokenTaken);
//if ($this->debugLevel) printf("<b>function</b> %s::lexical tokenTaken=[$tokenTaken] lowerToken=[$lowerToken]<br>\n", get_class($this));

				switch ($lowerToken)
				{
				// Actions
				case 'select':
				case 'update':
				case 'insert':
				case 'delete':
				case 'create':
					$lexType = 'A';
					$tokenTaken = $lowerToken;
					break;

				// Keywords
				case 'where':
				case 'into':
				case 'from':
				case 'set':
				case 'values':
				case 'order':
				case 'by':
				case 'desc':
				case 'table':
				case 'numeric':
				case 'char':
				case 'auto_increment':
				case 'distinct':
					$lexType = 'K';
					$tokenTaken = $lowerToken;
					break;

				// Comparisons
				case '=':
				case '<>':
				case '>':
				case '>=':
				case '<':
				case '<=':
				case 'like':	// Same as =
				case 'rlike':	// Same as regexp
				case 'regexp':	// Regular Expression match
					$lexType = 'C';
					break;

				// Relations
				case 'and':
				case 'or':
					$lexType = 'R';
					$tokenTaken = $lowerToken;
					break;

				// Punctuations
				case '(':
				case ')':
				case '{':
				case '}':
				case ',':
					$lexType = 'P';
					break;

				// Terminator
				case VIRT_TERMTOKEN:
					$lexType = VIRT_TERMTYPE;
					break;

				// Literal
				default:
					$lexType = 'L';
				}
//if ($this->debugLevel) printf("<b>function</b> %s::lexical after switch tokenTaken=[$tokenTaken]<br>\n", get_class($this));
			}

			if ($lexType == 'L')
			{
				// Check for quote
				if ($tokenTaken[0] == "'")
				{
//if ($this->debugLevel) printf("<b>function</b> %s::lexical quote starts<br>\n", get_class($this));
					if (strlen($tokenTaken) == 1
					|| ($tokenTaken[strlen($tokenTaken) - 1] != "'"
					|| $tokenTaken[strlen($tokenTaken) - 2] == '\\'))
					{
						// A false end quote; no lexType; keep going
						$litSeg = $tokenTaken;
						$lexType = '';
//if ($this->debugLevel) printf("<b>function</b> %s::lexical litSeg=[$litSeg]<br>\n", get_class($this));
					} else
					{
						// End of quote
						// Dirty fix for " ," back to "," (genuine " ," will be replaced too)
						$ttsaved = $tokenTaken;
						$tokenTaken = ereg_replace(" \\\\,", "\\,", $tokenTaken);
					//	$tokenTaken = ereg_replace("^'", "", $tokenTaken);
					//	$tokenTaken = ereg_replace("'$", "", $tokenTaken);
					}
				}
			}

			if ($lexType)
			{
				$this->lexArr[] = array('type' => $lexType, 'token' => $tokenTaken);
//if ($this->debugLevel) printf("<b>function</b> %s::lexical pushed [$lexType-<b>$tokenTaken</b>]<br>\n", get_class($this));
				//if ($gotComma)
				//{
				//	// Prepare pushing a comma
				//	$lexType = 'P';
				//	$tokenTaken = ',';
				//	$this->lexArr[] = array('type' => $lexType, 'token' => $tokenTaken);
//if ($this->debugLevel) printf("<b>function</b> %s::lexical pushed [$lexType-<b>$tokenTaken</b>]<br>\n", get_class($this));
				//}
			}
		}

		$this->lexArr[] = array('type' => VIRT_TERMTYPE, 'token' => VIRT_TERMTOKEN);  // One more just in case
		return $this->lexArr;
	}


	function matchToken($lex, $type, $token = '')
	{
		return ($lex['type'] == $type && ($token == '' || $lex['token'] == $token));
	}


	function isTerminator($lex)
	{
		return $this->matchToken($lex, VIRT_TERMTYPE, VIRT_TERMTOKEN);
	}


	function lastParsed()
	{
		return current($this->lexArr);
	}


	function lexicalText($lex = '')
	{
		if (empty($lex)) $lex = current($this->lexArr);
		return $lex['type'] . '-<b>' . $lex['token'] . '</b>';
	}


/*
Parsing algorithms:

General Note:
	After parsing a construct, the lexArr pointer point to the next
	token immediately after the construct.

Compulsory construct:
	if (DB::isError($ret = $this->parse...())) return $ret;

Single Token without processing:
	if (!$this->matchToken(next($this->lexArr), ...)) return $this->raiseError(DB_ERROR_SYNTAX);

Single Token:
	if (!$this->matchToken(($lex = next($this->lexArr)), ...)) return $this->raiseError(DB_ERROR_SYNTAX);
	... $lex['token'];

Optional:
	if (!$this->matchToken(next($this->lexArr), ...start...)) prev($this->lexArr);
	else if (DB::isError($ret = $this->parse...())) return $ret;
	//Part-of-else-for-ending-if-exists: if (!$this->matchToken(next($this->lexArr), ...end...)) return $this->raiseError(DB_ERROR_SYNTAX);

Repetition (Optional): [Element [, Separator] ...]
	if ($this->matchToken(($lex = next($this->lexArr)), ...element...))
	{
		... $lex['token'];
		while ($this->matchToken(next($this->lexArr), ...separator...))
		{
			if (!$this->matchToken(($lex = next($this->lexArr)), ...element...)) return $this->raiseError(DB_ERROR_SYNTAX);
			... $lex['token'];
		}
	}
	prev($this->lexArr);

Repetition (Compulsory): Element [, Separator] ...
	do
	{
		if (!$this->matchToken(($lex = next($this->lexArr)), ...element...)) return $this->raiseError(DB_ERROR_SYNTAX);
		... $lex['token'];
	} while ($this->matchToken(next($this->lexArr), ...separator...))
	prev($this->lexArr);
 */


	function parse()
	{
		$this->sqlAction = '';
		$this->tableName = '';
		$this->fieldList = array();
		$this->valueList = array();
		$this->whereFuncBody = '';
		$this->orderByList = array();
		$this->distinct = false;

		reset($this->lexArr);
//if ($this->debugLevel) printf("<b>function</b> %s::parse at [%s]<br>\n", get_class($this), $this->lexicalText());

		if (!$this->matchToken(($lex = current($this->lexArr)), 'A')) return $this->raiseError(DB_ERROR_SYNTAX);
		$this->sqlAction = $lex['token'];

//if ($this->debugLevel) printf("<b>function</b> %s::parse sqlAction=[%s]<br>\n", get_class($this), $this->sqlAction);

		$ret = 0;
		switch ($this->sqlAction)
		{
		case 'select':
			$ret = $this->parseSelect();
			break;
		case 'update':
			$ret = $this->parseUpdate();
			break;
		case 'insert':
			$ret = $this->parseInsert();
			break;
		case 'delete':
			$ret = $this->parseDelete();
			break;
		case 'create':
			$ret = $this->parseCreate();
			break;
		default:
			return $this->raiseError(DB_ERROR_SYNTAX);
		}

		if (!$this->isTerminator(next($this->lexArr))) return $this->raiseError(DB_ERROR_SYNTAX);

		return $ret;
	}


	function parseSelect()
	{
if ($this->debugLevel) printf("<b>function</b> %s::parseSelect after [%s]<br>\n", get_class($this), $this->lexicalText());
// distinct
		if (!$this->matchToken(next($this->lexArr), 'K', 'distinct')) prev($this->lexArr);
		else $this->distinct = true;
// field list
		if (DB::isError($ret = $this->parseList($this->fieldList, true))) return $ret;
// from
		if (!$this->matchToken(next($this->lexArr), 'K', 'from')) return $this->raiseError(DB_ERROR_SYNTAX);
// tableName
		if (!$this->matchToken(($lex = next($this->lexArr)), 'L')) return $this->raiseError(DB_ERROR_SYNTAX);
		$this->tableName = $lex['token'];
// where
		if (!$this->matchToken(next($this->lexArr), 'K', 'where')) prev($this->lexArr);
		else if (DB::isError($ret = $this->parseWhereClause())) return $ret;
// order
		if (!$this->matchToken(next($this->lexArr), 'K', 'order')) prev($this->lexArr);
		else if (DB::isError($ret = $this->parseOrderByClause())) return $ret;

		return DB_OK;
	}


	function parseUpdate()
	{
if ($this->debugLevel) printf("<b>function</b> %s::parseUpdate after [%s]<br>\n", get_class($this), $this->lexicalText());
// tableName
		if (!$this->matchToken(($lex = next($this->lexArr)), 'L')) return $this->raiseError(DB_ERROR_SYNTAX);
		$this->tableName = $lex['token'];
// set
		if (!$this->matchToken(next($this->lexArr), 'K', 'set')) return $this->raiseError(DB_ERROR_SYNTAX);
// set list
		if (DB::isError($ret = $this->parseSetList())) return $ret;
// where
		if (!$this->matchToken(next($this->lexArr), 'K', 'where')) prev($this->lexArr);
		else if (DB::isError($ret = $this->parseWhereClause())) return $ret;

		return DB_OK;
	}


	function parseInsert()
	{
if ($this->debugLevel) printf("<b>function</b> %s::parseInsert after [%s]<br>\n", get_class($this), $this->lexicalText());
// into
		if (!$this->matchToken(next($this->lexArr), 'K', 'into')) return $this->raiseError(DB_ERROR_SYNTAX);
// tableName
		if (!$this->matchToken(($lex = next($this->lexArr)), 'L')) return $this->raiseError(DB_ERROR_SYNTAX);
		$this->tableName = $lex['token'];
// ( fieldlist )
		if (!$this->matchToken(next($this->lexArr), 'P', '('))
		{
			prev($this->lexArr);
		} else
		{
			if (DB::isError($ret = $this->parseList($this->fieldList, true))) return $ret;
if ($this->debugLevel)
{
	printf("<b>function</b> %s::parseInsert got fieldList:<br>\n", get_class($this));
	print_r($this->fieldList);
	printf("<br>\n");
}
			if (!$this->matchToken(next($this->lexArr), 'P', ')')) return $this->raiseError(DB_ERROR_SYNTAX);
		}
// values
		if (!$this->matchToken(next($this->lexArr), 'K', 'values')) return $this->raiseError(DB_ERROR_SYNTAX);
		if (!$this->matchToken(next($this->lexArr), 'P', '(')) return $this->raiseError(DB_ERROR_SYNTAX);
		if (DB::isError($ret = $this->parseList($this->valueList))) return $ret;
		if (!$this->matchToken(next($this->lexArr), 'P', ')')) return $this->raiseError(DB_ERROR_SYNTAX);

		return DB_OK;
	}


	function parseDelete()
	{
if ($this->debugLevel) printf("<b>function</b> %s::parseDelete after [%s]<br>\n", get_class($this), $this->lexicalText());
// from
		if (!$this->matchToken(next($this->lexArr), 'K', 'from')) return $this->raiseError(DB_ERROR_SYNTAX);
// tableName
		if (!$this->matchToken(($lex = next($this->lexArr)), 'L')) return $this->raiseError(DB_ERROR_SYNTAX);
		$this->tableName = $lex['token'];
// where
		if (!$this->matchToken(next($this->lexArr), 'K', 'where')) prev($this->lexArr);
		else if (DB::isError($ret = $this->parseWhereClause())) return $ret;

		return DB_OK;
	}


	function parseCreate()
	{
if ($this->debugLevel) printf("<b>function</b> %s::parseCreate after [%s]<br>\n", get_class($this), $this->lexicalText());
// table
		if (!$this->matchToken(next($this->lexArr), 'K', 'table')) return $this->raiseError(DB_ERROR_SYNTAX);
// tableName
		if (!$this->matchToken(($lex = next($this->lexArr)), 'L')) return $this->raiseError(DB_ERROR_SYNTAX);
		$tbl = $lex['token'];
		$this->tableName = $tbl;
		$this->tableFields[$tbl] = array();
		$this->tableTypes[$tbl] = array();
		$this->tableAutoInc[$tbl] = '';
		$this->tableLastIns[$tbl] = 0;
		$this->tableRecords[$tbl] = array();
// { creation_list }
		if (!$this->matchToken(next($this->lexArr), 'P', '(')) return $this->raiseError(DB_ERROR_SYNTAX);
		do
		{
			if (DB::isError($ret = $this->parseCreateField())) return $ret;
		} while ($this->matchToken(($lex = next($this->lexArr)), 'P', ','));
		if (!$this->matchToken($lex, 'P', ')')) return $this->raiseError(DB_ERROR_SYNTAX);

		return DB_OK;
	}


	function parseCreateField()
	{
if ($this->debugLevel) printf("<b>function</b> %s::parseCreateField after [%s]<br>\n", get_class($this), $this->lexicalText());
		if (!$this->matchToken(($lex = next($this->lexArr)), 'L')) return $this->raiseError(DB_ERROR_SYNTAX);
		$fld = $lex['token'];
		$this->tableFields[$this->tableName][] = $fld;
		$type = VIRT_CHAR;
		while ($this->matchToken(($lex = next($this->lexArr)), 'K'))
		{
			switch ($lex['token'])
			{
			case 'char':
				$type = VIRT_CHAR;
				break;
			case 'numeric':
				$type = VIRT_NUMERIC;
				break;
			case 'auto_increment':
				$this->tableAutoInc[$this->tableName] = $fld;
				$type = VIRT_NUMERIC;  // Must be numeric in order to be auto_increment
				break;
			default:
				return $this->raiseError(DB_ERROR_SYNTAX);
			}
		}

		$this->tableTypes[$this->tableName][$fld] = $type;
		prev($this->lexArr);
		return DB_OK;
	}


	function parseList(&$target, $isFldLst = false)
	{
if ($this->debugLevel) printf("<b>function</b> %s::parseList after [%s]<br>\n", get_class($this), $this->lexicalText());
		if ($this->matchToken(($lex = next($this->lexArr)), 'L'))
		{
			$target[] = $lex['token'];
if ($this->debugLevel) printf("<b>function</b> %s::parseList pushed [%s]<br>\n", get_class($this), $lex['token']);
			while ($this->matchToken(next($this->lexArr), 'P', ','))
			{
				if (!$this->matchToken(($lex = next($this->lexArr)), 'L')) return $this->raiseError(DB_ERROR_SYNTAX);
				$target[] = $lex['token'];
if ($this->debugLevel) printf("<b>function</b> %s::parseList pushed [%s]<br>\n", get_class($this), $lex['token']);
			}
		}
		prev($this->lexArr);
		if ($isFldLst) foreach ($target as $key => $val) $target[$key] = ereg_replace('^.*\.', '', $val);

		return DB_OK;
	}


	function parseWhereBrackets()
	{
		$lex = next($this->lexArr);
		while ($this->matchToken($lex, 'P', '(') || $this->matchToken($lex, 'P', ')'))
		{
			$this->whereFuncBody .= ' ' . $lex['token'];
			$lex = next($this->lexArr);
		}
		prev($this->lexArr);

		return DB_OK;
	}


	function parseWhereCondition()
	{
if ($this->debugLevel) printf("<b>function</b> %s::parseWhereCondition after [%s]<br>\n", get_class($this), $this->lexicalText());
		$this->parseWhereBrackets();

		$whereFuncBodySaved = $this->whereFuncBody;
		$this->whereFuncBody .= ' (';

		if (!$this->matchToken(($lex = next($this->lexArr)), 'L')) return $this->raiseError(DB_ERROR_SYNTAX);
		//$this->whereFuncBody .= ' $rec[\'' . $lex['token'] . '\']';
		//$this->whereFuncBody .= ' $rec[\'' . ereg_replace('^.*\.', '', $lex['token']) . '\']';
		$lhs = '$rec[\'' . ereg_replace('^.*\.', '', $lex['token']) . '\']';
		$this->whereFuncBody .= " $lhs";

		if (!$this->matchToken(($lex = next($this->lexArr)), 'C')) return $this->raiseError(DB_ERROR_SYNTAX);
		$op = $lex['token'];
		switch ($op)
		{
		case '=':
		case 'like':
			$op = '==';
			break;
		case '<>':
			$op = '!=';
			break;
		}
		$this->whereFuncBody .= ' ' . $op;

		if (!$this->matchToken(($lex = next($this->lexArr)), 'L')) return $this->raiseError(DB_ERROR_SYNTAX);
		//$this->whereFuncBody .= ' ' . $lex['token'];
		$rhs = $lex['token'];
		$this->whereFuncBody .= ' ' . $rhs;

		$this->whereFuncBody .= ' )';

		if ($op == 'rlike')
		{
			// Major surgery for whereFuncBody
			$this->whereFuncBody = $whereFuncBodySaved . ' ereg( ' . sql_regcase($rhs) . ' , ' . $lhs . ' )';
		}

		$this->parseWhereBrackets();

		return DB_OK;
	}


	function parseWhereClause()
	{
if ($this->debugLevel) printf("<b>function</b> %s::parseWhereClause after [%s]<br>\n", get_class($this), $this->lexicalText());
		// The first element is Compulsory
		if (DB::isError($ret = $this->parseWhereCondition())) return $ret;
		while ($this->matchToken(($lex = next($this->lexArr)), 'R'))
		{
			$op = $lex['token'];
			switch ($op)
			{
			case 'and':
				$op = '&&';
				break;
			case 'or':
				$op = '||';
				break;
			}
			$this->whereFuncBody .= ' ' . $op;
		
			if (DB::isError($ret = $this->parseWhereCondition())) return $ret;
		}
		prev($this->lexArr);

		return DB_OK;
	}


	function parseOrderByClause()
	{
if ($this->debugLevel) printf("<b>function</b> %s::parseOrderByClause after [%s]<br>\n", get_class($this), $this->lexicalText());
// by
		if (!$this->matchToken(next($this->lexArr), 'K', 'by')) return $this->raiseError(DB_ERROR_SYNTAX);
// fld [desc], ...
		do
		{
			if (!$this->matchToken(($lex = next($this->lexArr)), 'L')) return $this->raiseError(DB_ERROR_SYNTAX);
			$ordFld = $lex['token'];
			$ordDesc = SORT_ASC;
			$ordType = SORT_REGULAR;
			if (!$this->matchToken(next($this->lexArr), 'K', 'desc')) prev($this->lexArr);
			else $ordDesc = SORT_DESC;
if ($this->debugLevel) printf("<b>function</b> %s::parseOrderByClause ordType based on [%s]<br>\n", get_class($this), $this->tableTypes[$this->tableName][$ordFld]);
			if ($this->tableTypes[$this->tableName][$ordFld] == VIRT_NUMERIC) $ordType = SORT_NUMERIC;
			$this->orderByList[] = array('field' => $ordFld, 'desc' => $ordDesc, 'type' => $ordType);
		} while ($this->matchToken(next($this->lexArr), 'P', ','));
		prev($this->lexArr);

		return DB_OK;
	}


	function parseFieldValuePair()
	{
if ($this->debugLevel) printf("<b>function</b> %s::parseFieldValuePair after [%s]<br>\n", get_class($this), $this->lexicalText());
		if (!$this->matchToken(($lex = next($this->lexArr)), 'L')) return $this->raiseError(DB_ERROR_SYNTAX);
		$this->fieldList[] = ereg_replace('^.*\.', '', $lex['token']);

		if (!$this->matchToken(($lex = next($this->lexArr)), 'C', '=')) return $this->raiseError(DB_ERROR_SYNTAX);

		if (!$this->matchToken(($lex = next($this->lexArr)), 'L')) return $this->raiseError(DB_ERROR_SYNTAX);
		$this->valueList[] = $lex['token'];

		return DB_OK;
	}


	function parseSetList()
	{
if ($this->debugLevel) printf("<b>function</b> %s::parseSetList after [%s]<br>\n", get_class($this), $this->lexicalText());
		// The first element is Compulsory
		do
		{
			if (DB::isError($ret = $this->parseFieldValuePair())) return $ret;
		} while ($this->matchToken(next($this->lexArr), 'P', ','));
		prev($this->lexArr);

		return DB_OK;
	}


/************************************************************
 *
 * aux functions
 *
 */

	function getLexicalArray()
	{
		return $this->lexArr;
	}


	function getSqlAction()
	{
		return $this->sqlAction;
	}


	function tableExists($tblName)
	{
		return isset($this->tableFields[$tblName]);
	}


	function getTableFields()
	{
		return $this->tableFields;
	}


	function getTableTypes()
	{
		return $this->tableTypes;
	}


	function getTableAutoInc()
	{
		return $this->tableAutoInc;
	}


	function getTableLastIns()
	{
		return $this->tableLastIns;
	}


	function getTableRecords()
	{
		return $this->tableRecords;
	}


	function getTableName()
	{
		return $this->tableName;
	}


	function getFieldList()
	{
		return $this->fieldList;
	}


	function getValueList()
	{
		return $this->valueList;
	}


	function getWhereFuncBody()
	{
		return $this->whereFuncBody;
	}


	function getOrderByList()
	{
		return $this->orderByList;
	}


	function showCreateTable()
	{
		$ret = '';
		if (!is_array($this->tableFields)) return $ret;
		foreach($this->tableFields as $tbl => $fldlst)
		{
			$ret .= "create table $tbl\n(\n";
			$isFirst = true;
			foreach($fldlst as $fld)
			{
				if (!$isFirst) $ret .= ",\n";
				$isFirst = false;
				$ret .= "\t$fld";
				switch($this->tableTypes[$tbl][$fld])
				{
				case VIRT_CHAR:
					$ret .= " char";
					break;
				case VIRT_NUMERIC:
					$ret .= " numeric";
					break;
				}
				if ($this->tableAutoInc[$tbl] == $fld) $ret .= " auto_increment";
			}
			$ret .= "\n)";
		}
		return $ret;
	}


	function showInsertInto()
	{
		$ret = array();
		if (!is_array($this->tableRecords[$this->tableName])) return $ret;
		foreach($this->tableRecords[$this->tableName] as $recKey => $rec)
		{
			$ln = sprintf("insert into %s values (", $this->tableName);
			//foreach($rec as $val) $ln .= " '$val',";
			foreach($this->tableFields[$this->tableName] as $fld)
			{
				$fldVal = $rec[$fld];
				$fldVal = ereg_replace(",", "\\,", $fldVal);
				$fldVal = ereg_replace("\'", "\\'", $fldVal);
				$ln .= " '" . $fldVal . "',";
			}
			$ln = ereg_replace(',$', " )", $ln);
			$ret[] = $ln;
if ($this->debugLevel) printf("<b>function</b> %s::showInsertInto: [%s]<br>\n", get_class($this), $ln);
		}
		return $ret;
	}


	// For debugging only
	function dumpTables()
	{
		foreach($this->tableFields as $tbl => $fldlst)
		{
			printf("<font color=#0000ff>create table <b>$tbl</b> ( ");
			$isFirst = true;
			foreach($fldlst as $fld)
			{
				if (!$isFirst) printf(", ");
				$isFirst = false;
				printf(" <b>$fld</b> ");
				switch($this->tableTypes[$tbl][$fld])
				{
				case VIRT_CHAR:
					printf("char");
					break;
				case VIRT_NUMERIC:
					printf("numeric");
					break;
				}
				if ($this->tableAutoInc[$tbl] == $fld) printf(" auto_increment");
			}
			printf(" )</font><br>\n");
		}
	}


	// Get the record set in $result from another database object to this object
	function importResult($query, $dbobj, $result)
	{
		$qry = $this->prepare($query);
if ($this->debugLevel) printf("<b>function</b> %s::importResult prepare resturns [$qry]<br>\n", get_class($this), $qry);
		//while (($row = $dbobj->fetchRow($result)) && count($row) > 0)
		$dbRes = new DB_result($dbobj, $result);
		while (($row = $dbRes->fetchRow()) && count($row) > 0)
		{
if ($this->debugLevel)
{
	printf("<b>function</b> %s::importResult got row: ", get_class($this));
	print_r($row);
	printf("<br>\n");
}
			if (DB::isError($res = $this->execute($qry, $row))) return $res;
		}
		return DB_OK;
	}


	// Get the record set in $result from this object to another database object
	function exportResult($query, $dbobj, $result)
	{
		$qry = $dbobj->prepare($query);
if ($this->debugLevel) printf("<b>function</b> %s::exportResult prepare resturns [$qry]<br>\n", get_class($this), $qry);
		while (($row = $this->virt_fetch_row($result)) && count($row) > 0)
		{
if ($this->debugLevel)
{
	printf("<b>function</b> %s::exportResult got row: ", get_class($this));
	print_r($row);
	printf("<br>\n");
}
			if (DB::isError($res = $dbobj->execute($qry, $row))) return $res;
		}
		return DB_OK;
	}


	// Get the record set from reading commands from a file descriptor
	function load($filename)
	{
if ($this->debugLevel) printf("<b>function</b> %s::load<br>\n", get_class($this));
		$fd = fopen($filename, 'r');
		if ($fd === false) return false;
if ($this->debugLevel) printf("<b>function</b> %s::load: %s opened<br>\n", get_class($this), $filename);
		$ln = '';
		while ($lnread = fgets($fd, VIRT_MAXLINE))
		{
			$lnread = ereg_replace('^#.*', '', $lnread);  // No comments
			$lnread = ereg_replace('^[	]*', '', $lnread);  // No leading spaces
			$lnread = ereg_replace("[	]*\n$", '', $lnread);  // No trailing white spaces
if ($this->debugLevel) printf("<b>function</b> %s::load: lnread=[%s]<br>\n", get_class($this), $lnread);
			if (empty($lnread)) continue;
			$ln .= (($ln != '') ? ' ' : '') . $lnread;
if ($this->debugLevel) printf("<b>function</b> %s::load: ln=[%s]<br>\n", get_class($this), $ln);
			if (!ereg(VIRT_TERMTOKEN . '$', $ln)) continue;  // No Terminator; to be continued

if ($this->debugLevel) printf("<b>function</b> %s::load: Process [%s]<br>\n", get_class($this), $ln);
			$res = $this->simpleQuery($ln);
			if (DB::isError($res))
			{
if ($this->debugLevel) printf("<b>function</b> %s::load: Got error [%s]<br>\n", get_class($this), DB::errorMessage($res));
				fclose($fd);
				return false;
			}

			$ln = '';
		}
		fclose($fd);

		$this->loadedFileName = $filename;

if ($this->debugLevel) var_dump($this);
		return true;
	}


	// Get the record set in $result from this object to another database object
	function unload($filenameArg = '')
	{
		if (!is_array($this->tableRecords[$this->tableName])) return false;

if ($this->debugLevel) printf("<b>function</b> %s::unload<br>\n", get_class($this));
		$filename = $filenameArg;
		if ($filename == '') $filename = $this->loadedFileName;
		$fd = @fopen($filename, 'w');
		if ($fd === false) return false;

		// Must not use EOD as it will create "\r\n" for line breaks on Windows,
		// which will jeopardise the loading and parsing of the file.
		fputs($fd, sprintf("#\n# Table structure for table '%s'\n#\n\n%s;\n\n"
, $this->tableName, $this->showCreateTable()));
		$insertInto = $this->showInsertInto();
if ($this->debugLevel) print_r($insertInto);
		foreach($insertInto as $iiQry) fputs($fd, sprintf("%s;\n", $iiQry));
		fputs($fd, "\n");
		fclose($fd);
		return true;
	}


	function setDebugLevel($level)
	{
//if ($this->debugLevel || $level) printf("<b>function</b> %s::setDebugLevel([$level])<br>\n", get_class($this));
		$this->debugLevel = $level;
		return DB_OK;
	}



/************************************************************
 *
 * virt database simulation functions
 *
 */


	function virt_query($query, $conn = 0)
	{
if ($this->debugLevel)
{
	printf("<b>function</b> %s::virt_query: query is<br>\n", get_class($this));
	var_dump($query);
	printf("<br>\n");
}
		$this->virtErrorNum = DB_OK;

		$this->lexical($query);
		if (DB::isError($this->parse()))
		{
			$this->virtErrorNum = DB_ERROR_SYNTAX;
			return false;
		}

		$res = 0;
		switch ($this->sqlAction)
		{
		case 'select':
			$res = $this->virt_query_select();
			break;
		case 'update':
			$res = $this->virt_query_update();
			break;
		case 'insert':
			$res = $this->virt_query_insert();
			break;
		case 'delete':
			$res = $this->virt_query_delete();
			break;
		case 'create':
			$res = true;  // Done already in parse
			break;
		}

		return $res;
	}


	function virt_query_select()
	{
if ($this->debugLevel) printf("<b>function</b> %s::virt_query_select<br>\n", get_class($this));
		if ($this->fieldList[0] == '*') $this->fieldList = $this->tableFields[$this->tableName];

		$selected = array();
		$distinctStrings = array();
		if (empty($this->whereFuncBody)) $this->whereFuncBody = ' true';
if ($this->debugLevel) printf("<b>function</b> %s::virt_query_select got whereFuncBody=[%s]<br>\n", get_class($this), $this->whereFuncBody);
		$whereFunc = @create_function('$rec', 'return ' . $this->whereFuncBody . ';');
		if (isset($this->tableRecords[$this->tableName]))
		{
			foreach($this->tableRecords[$this->tableName] as $recKey => $rec)
			{
				if ($whereFunc)
				{
					if ($whereFunc($rec))
					{
						$selected[] = $recKey;
						if ($this->distinct)
						{
							//$distStr = array();
							$distStr = '';
							foreach($this->fieldList as $fld) $distStr .= $rec[$fld] . '|';  //$distStr[] = $rec[$fld];
							$distinctStrings[$recKey] = $distStr;
						}
					}
				}
			}
		}

		// Distinct
		if ($this->distinct) $selected = array_keys(array_unique($distinctStrings));

		// Order By
		if (count($this->orderByList) > 0)
		{
			// Build an array of column values for every field in the orderByList
			// and put these arrays into sortKey

			$sortKey = array();
			foreach($this->orderByList as $oblkey => $ob) $sortKey[$oblkey] = array();

			foreach($selected as $recKey)
			{
				$rec = $this->tableRecords[$this->tableName][$recKey];
				// Go through selected and collect values from the tableRecords
				foreach($this->orderByList as $oblkey => $ob) $sortKey[$oblkey][] = $rec[$ob['field']];
			}

			// The following is clumsy as I couldn't figure out how to
			// make array_multisort take an array of array as argument.
if ($this->debugLevel)
{
	printf("SORT_ASC = %d<br>\n", SORT_ASC);
	printf("SORT_DESC = %d<br>\n", SORT_DESC);
	printf("SORT_REGULAR = %d<br>\n", SORT_REGULAR);
	printf("SORT_NUMERIC = %d<br>\n", SORT_NUMERIC);
	printf("SORT_STRING = %d<br>\n", SORT_STRING);
	printf("<b>function</b> %s::virt_query_select sorting ...<br>\n", get_class($this));
	printf("array_multisort(<br>\n");
	foreach($this->orderByList as $oblkey => $ob)
	{
		printf("sortKey[$oblkey], %d, %d,<br>\n", $ob['desc'], $ob['type']);
	}
	printf("selected)<br>\n");
}
			switch (count($this->orderByList))
			{
			case 1:
				array_multisort(
					$sortKey[0], $this->orderByList[0]['desc'], $this->orderByList[0]['type'],
					$selected);
				break;
			case 2:
				array_multisort(
					$sortKey[0], $this->orderByList[0]['desc'], $this->orderByList[0]['type'],
					$sortKey[1], $this->orderByList[1]['desc'], $this->orderByList[1]['type'],
					$selected);
				break;
			case 3:
				array_multisort(
					$sortKey[0], $this->orderByList[0]['desc'], $this->orderByList[0]['type'],
					$sortKey[1], $this->orderByList[1]['desc'], $this->orderByList[1]['type'],
					$sortKey[2], $this->orderByList[2]['desc'], $this->orderByList[2]['type'],
					$selected);
				break;
			default:  // At most 4
				array_multisort(
					$sortKey[0], $this->orderByList[0]['desc'], $this->orderByList[0]['type'],
					$sortKey[1], $this->orderByList[1]['desc'], $this->orderByList[1]['type'],
					$sortKey[2], $this->orderByList[2]['desc'], $this->orderByList[2]['type'],
					$sortKey[3], $this->orderByList[3]['desc'], $this->orderByList[3]['type'],
					$selected);
				break;
			}
		}

		reset($selected);
		$this->resultArray[] = array('tableName' => $this->tableName, 'fieldList' => $this->fieldList, 'resultRecords' => $selected);
		end($this->resultArray);
		$result = key($this->resultArray);
if ($this->debugLevel)
{
	printf("<b>function</b> %s::virt_query_select returns [$result] => an array of %d.<br>\n", get_class($this), count($this->resultArray[$result]['resultRecords']));
	print_r($this->resultArray[$result]['resultRecords']);
	printf("<br>\n");
}
//if (getServerVar('REMOTE_ADDR') == '220.240.52.238') $this->debugLevel = 0;
		return $result;
	}


	function virt_query_update()
	{
if ($this->debugLevel) printf("<b>function</b> %s::virt_query_update<br>\n", get_class($this));
		$result = $this->virt_query_select();
		if (!$result) return false;

		$tbl = $this->resultArray[$result]['tableName'];
if ($this->debugLevel) printf("<b>function</b> %s::virt_query_update got tbl=[$tbl]<br>\n", get_class($this));
		$fldLst = $this->resultArray[$result]['fieldList'];
		$valLst = $this->valueList;

		foreach($this->resultArray[$result]['resultRecords'] as $rec)
		{
if ($this->debugLevel) printf("<b>function</b> %s::virt_query_update on $tbl record #$rec<br>\n", get_class($this));
			reset($fldLst);
			reset($valLst);

			$fld = current($fldLst);
			$val = deQuote(current($valLst));

			//while ($fld && $val)
			while ($fld)
			{
if ($this->debugLevel) printf("<b>function</b> %s::virt_query_update on field $fld [%s] -> [%s]<br>\n", get_class($this), $this->tableRecords[$tbl][$rec][$fld], $val);
				$this->tableRecords[$tbl][$rec][$fld] = $val;
				$fld = next($fldLst);
				$val = deQuote(next($valLst));
			}
		}
		//reset($this->resultArray[$result]['resultRecords']);
		$this->freeResult($result);
		return true;
	}


	function virt_query_insert()
	{
if ($this->debugLevel) printf("<b>function</b> %s::virt_query_insert<br>\n", get_class($this));
		if (count($this->fieldList) == 0) $this->fieldList = $this->tableFields[$this->tableName];
if ($this->debugLevel)
{
	printf("<b>function</b> %s::virt_query_insert got fieldList:<br>\n", get_class($this));
	print_r($this->fieldList);
	printf("<br>\n");
}
		$autoIncFld = $this->tableAutoInc[$this->tableName];
		if (!empty($autoIncFld) && !in_array($autoIncFld, $this->fieldList))
		{
			array_push($this->fieldList, $autoIncFld);
			array_push($this->valueList, $this->tableLastIns[$this->tableName] + 1);
		}
		$newrec = mergeKeyValue($this->fieldList, $this->valueList);
		foreach($newrec as $key => $val) $newrec[$key] = deQuote(unEscape($val));
		if (!empty($autoIncFld))
		{
			$this->tableLastIns[$this->tableName] = $newrec[$autoIncFld];
			$this->lastInsertID = $this->tableLastIns[$this->tableName];
		}
		$this->tableRecords[$this->tableName][] = $newrec;
		return true;
	}


	function virt_query_delete()
	{
if ($this->debugLevel) printf("<b>function</b> %s::virt_query_delete<br>\n", get_class($this));
		$result = $this->virt_query_select();
		if (!$result) return false;

		$tbl = $this->resultArray[$result]['tableName'];

		foreach($this->resultArray[$result]['resultRecords'] as $rec)
		{
			unset($this->tableRecords[$tbl][$rec]);
		}

		$this->freeResult($result);
		return true;
	}


	function virt_errno($conn = 0)
	{
		return $this->virtErrorNum;
	}


	function virt_fetch_array($result, $fetchmode = VIRT_BOTH)
	{
if ($this->debugLevel) printf("<b>function</b> %s::virt_fetch_array([$result], [$fetchmode])<br>\n", get_class($this));
		if (DB::isError($result) || $result == 0) return false;
		$tbl = $this->resultArray[$result]['tableName'];
		$rec = current($this->resultArray[$result]['resultRecords']);
		if ($rec === false)
		{
			//$this->virtErrorNum = DB_ERROR;
			//return false;
			return array();
		}
		next($this->resultArray[$result]['resultRecords']);

		$fetchAssoc = true;
		$fetchNum = true;
		switch ($fetchmode)
		{
		case VIRT_ASSOC:
			$fetchNum = false;
			break;
		case VIRT_NUM:
			$fetchAssoc = false;
			break;
		}

		$row = array();

		// Do fetchNum first as it will give a natural sequence that it requires.
		if ($fetchNum) foreach($this->resultArray[$result]['fieldList'] as $fld)
		{
			$row[] = $this->tableRecords[$tbl][$rec][$fld];
		}
		if ($fetchAssoc) foreach($this->resultArray[$result]['fieldList'] as $fld)
		{
			$row[$fld] = $this->tableRecords[$tbl][$rec][$fld];
if ($this->debugLevel) printf("<b>function</b> %s::virt_fetch_array: \$row[%s] got [%s]<br>\n", get_class($this), $fld, $row[$fld]);
		}

		return $row;
	}


	function virt_fetch_row($result)
	{
if ($this->debugLevel) printf("<b>function</b> %s::virt_fetch_row([$result])<br>\n", get_class($this));
		return $this->virt_fetch_array($result, VIRT_NUM);
	}


	function virt_free_result($result)
	{
if ($this->debugLevel) printf("<b>function</b> %s::virt_free_result([$result])<br>\n", get_class($this));
		if (!DB::isError($result) && $result > 1)
		{
			unset($this->resultArray[$result]);
if ($this->debugLevel) printf("<b>function</b> %s::virt_free_result unset<br>\n", get_class($this));
		}
		//unset($this->resultArray[$result]['fieldList']);
		//unset($this->resultArray[$result]['resultRecords']);
	}


	function virt_num_fields($result)
	{
		if (!isset($this->resultArray[$result])) return false;
		return count($this->resultArray[$result]['fieldList']);
	}


	function virt_num_rows($result)
	{
if ($this->debugLevel) printf("<b>function</b> %s::virt_num_rows($result) resultArray isset = [%d]<br>\n", get_class($this), isset($this->resultArray[$result]));
		if (!isset($this->resultArray[$result])) return false;
		return count($this->resultArray[$result]['resultRecords']);
	}


	function virt_data_seek($result, $rownum)
	{
//if ($this->debugLevel) printf("<b>function</b> %s::virt_data_seek rownum=[$rownum]<br>\n", get_class($this));
		if (!isset($this->resultArray[$result])) return false;

//if ($this->debugLevel) printf("this->resultArray[$result]['resultRecords']:\n");
//if ($this->debugLevel) print_r($this->resultArray[$result]['resultRecords']);
//if ($this->debugLevel) printf("<br>Seeking ... \n");
		reset($this->resultArray[$result]['resultRecords']);
		for ($i = 0; $i < $rownum; $i++)
		{
//if ($this->debugLevel) printf("$i:%d ", current($this->resultArray[$result]['resultRecords']));
			if (next($this->resultArray[$result]['resultRecords']) === false) return false;
		}
//if ($this->debugLevel) printf("<br>Stopped at $i:%d<br>\n", current($this->resultArray[$result]['resultRecords']));
		return true;
	}


	function virt_insert_id()
	{
		return $this->lastInsertID;
	}




/************************************************************
 *
 * DB object functions
 *
 */

// The following functions exist in other database objects as well.
// Avoid changing them as much as you can.  Only the virt_... functions
// need to be replaced by those defined before the class.

    // {{{ connect()

	/**
	 * Connect to a database and log in as the specified user.
	 *
	 * @param $dsn the data source name (see DB::parseDSN for syntax)
	 * @param $persistent (optional) whether the connection should
	 *        be persistent
	 * @access public
	 * @return int DB_OK on success, a DB error on failure
	 */
	/* Arguments are for compatibility only and will be ignored.
	function connect(&$dsn, $persistent = false) {
		if (is_array($dsn)) {
			$dsninfo = &$dsn;
		} else {
			$dsninfo = DB::parseDSN($dsn);
		}
		if (!$dsninfo || !$dsninfo['phptype']) {
			return $this->raiseError(); // XXX ERRORMSG
		}
		$dbhost = $dsninfo['hostspec'] ? $dsninfo['hostspec'] : 'localhost';
		$user = $dsninfo['username'];
		$pw = $dsninfo['password'];
		$connect_function = $persistent ? 'virt_pconnect' : 'virt_connect';
		if ($dbhost && $user && $pw) {
			$conn = $connect_function($dbhost, $user, $pw);
		} elseif ($dbhost && $user) {
			$conn = $connect_function($dbhost, $user);
		} elseif ($dbhost) {
			$conn = $connect_function($dbhost);
		} else {
			$conn = false;
		}
		if ($conn == false) {
			return $this->raiseError(); // XXX ERRORMSG
		}
		if ($dsninfo['database']) {
			if (!virt_select_db($dsninfo['database'], $conn)) {
				return $this->raiseError(); // XXX ERRORMSG
			}
		}
	 */
	function connect() {
		$conn = 0;

		$this->connection = $conn;
		return DB_OK;
	}

    // }}}
    // {{{ disconnect()

	/**
	 * Log out and disconnect from the database.
	 *
	 * @access public
	 *
	 * @return bool TRUE on success, FALSE if not connected.
	 */
	function disconnect() {
		//return virt_close($this->connection); // XXX ERRORMSG
		return DB_OK;
	}

    // }}}
    // {{{ query()


	/**
	 * Send a query to Virt and return the results as a DB_result object.
	 *
	 * @param $query the SQL query
	 *
	 * @access public
	 *
	 * @return object a DB_result object on success, a DB error
	 * on failure
	 */
	function &query($query) {
		$this->last_query = $query;
		$result = $this->virt_query($query, $this->connection);
		if (!$result) {
			return $this->raiseError($this->errorCode($this->virt_errno($this->connection)));
		}
		// Determine which queries that should return data, and which
		// should return an error code only.
		if (preg_match('/(SELECT|SHOW)/i', $query)) {
			$resultObj = new DB_result($this, $result);
			return $resultObj;
		} else {
			return DB_OK;
		}
	}

	// }}}
    // {{{ simpleQuery()

	/**
	 * Send a query to Virt and return the results as a Virt resource
	 * identifier.
	 *
	 * @param $query the SQL query
	 *
	 * @access public
	 *
	 * @return int returns a valid Virt result for successful SELECT
	 * queries, DB_OK for other successful queries.  A DB error is
	 * returned on failure.
	 */
	function simpleQuery($query) {
		$this->last_query = $query;
		$result = $this->virt_query($query, $this->connection);
		if (!$result) {
			return $this->raiseError($this->errorCode($this->virt_errno($this->connection)));
		}
		// Determine which queries that should return data, and which
		// should return an error code only.
		if (preg_match('/(SELECT|SHOW)/i', $query)) {
			return $result;
		} else {
			return DB_OK;
		}
	}

	// }}}
    // {{{ fetchRow()

	/**
	 * Fetch a row and return as array.
	 *
	 * @param $result Virt result identifier
	 * @param $fetchmode how the resulting array should be indexed
	 *
	 * @access public
	 *
	 * @return mixed an array on success, a DB error on failure, NULL
	 *               if there is no more data
	 OBSOLETE
	function &fetchRow($result, $fetchmode = DB_FETCHMODE_DEFAULT) {
		if ($fetchmode == DB_FETCHMODE_DEFAULT) {
			$fetchmode = $this->fetchmode;
		}
		if ($fetchmode & DB_FETCHMODE_ASSOC) {
			$row = $this->virt_fetch_array($result, VIRT_ASSOC);
		} else {
			$row = $this->virt_fetch_row($result);
		}
		if (!$row) {
			$errno = $this->virt_errno($this->connection);
			if (!$errno) {
				return NULL;
			}
			return $this->raiseError($this->errorCode($errno));
		}
		return $row;
	}
	 */

	// }}}
    // {{{ fetchInto()

	/**
	 * Fetch a row and insert the data into an existing array.
	 *
	 * @param $result Virt result identifier
	 * @param $arr (reference) array where data from the row is stored
	 * @param $fetchmode how the array data should be indexed
	 *
	 * @access public
	 *
	 * @return int DB_OK on success, a DB error on failure
	 */
	function fetchInto($result, &$arr, $fetchmode = DB_FETCHMODE_DEFAULT) {
		if ($fetchmode == DB_FETCHMODE_DEFAULT) {
			$fetchmode = $this->fetchmode;
		}
		if ($fetchmode & DB_FETCHMODE_ASSOC) {
			$arr = $this->virt_fetch_array($result, VIRT_ASSOC);
		} else {
			$arr = $this->virt_fetch_row($result);
		}
		if (!$arr) {
			$errno = $this->virt_errno($this->connection);
			if (!$errno) {
				return NULL;
			}
			return $this->raiseError($this->errorCode($errno));
		}
		return DB_OK;
	}

	// }}}
    // {{{ freeResult()

	/**
	 * Free the internal resources associated with $result.
	 *
	 * @param $result Virt result identifier or DB statement identifier
	 *
	 * @access public
	 *
	 * @return bool TRUE on success, FALSE if $result is invalid
	 */
	function freeResult($result) {
		//if (is_resource($result)) {
			return $this->virt_free_result($result);
		//}
		/*
		if (!isset($this->prepare_tokens[$result])) {
			return false;
		}
		unset($this->prepare_tokens[$result]);
		unset($this->prepare_types[$result]);
		 */
		//return true; 
	}

	// }}}
    // {{{ numCols()

	/**
	 * Get the number of columns in a result set.
	 *
	 * @param $result Virt result identifier
	 *
	 * @access public
	 *
	 * @return int the number of columns per row in $result
	 */
	function numCols($result) {
		$cols = $this->virt_num_fields($result);
		if (!$cols) {
			return $this->raiseError($this->errorCode($this->virt_errno($this->connection)));
		}
		return $cols;
	}

    // }}}
    // {{{ errorNative()

	/**
	 * Get the native error code of the last error (if any) that
	 * occured on the current connection.
	 *
	 * @access public
	 *
	 * @return int native Virt error code
	 */
	function errorNative() {
		return $this->virt_errno($this->connection);
	}

    // }}}
    // {{{ prepare()

	/**
	 * Prepares a query for multiple execution with execute().  With
	 * Virt, this is emulated.
	 *
	 * @param $query the SQL query to prepare for execution.  The
	 * characters "?" and "&" have special meanings. "?" is a scalar
	 * placeholder, the value provided for it when execute()ing is
	 * inserted as-is.  "&" is an opaque placeholder, the value
	 * provided on execute() is a file name, and the contents of that
	 * file will be used in the query.
	 *
	 * @access public
	 *
	 * @return int identifier for this prepared query
	 */
	function prepare($query) {
		$tokens = split('[\&\?]', $query);
		$token = 0;
		$types = array();
		for ($i = 0; $i < strlen($query); $i++) {
			switch ($query[$i]) {
				case '?':
					$types[$token++] = DB_PARAM_SCALAR;
					break;
				case '&':
					$types[$token++] = DB_PARAM_OPAQUE;
					break;
			}
		}
		$this->prepare_tokens[] = &$tokens;
		end($this->prepare_tokens);
		$k = key($this->prepare_tokens);
		$this->prepare_types[$k] = $types;
		return $k;
	}

    // }}}
    // {{{ execute()

	/**
	 * Executes a prepared query and substitutes placeholders with
	 * provided values.
	 *
	 * @return int returns a Virt result resource for successful
	 * SELECT queries, DB_OK for other successful queries.  A DB error
	 * is returned on failure.
	 *
	 * @see DB_virt::prepare
	 */
	function execute($stmt, $data = false) {
		$realquery = $this->execute_emulate_query($stmt, $data);
		$this->last_query = $realquery;
		$result = $this->virt_query($realquery, $this->connection);
		if (!$result) {
			return $this->raiseError($this->errorCode($this->virt_errno($this->connection)));
		}
		if (preg_match('/(SELECT|SHOW)/i', $realquery)) {
			return $result;
		} else {
			return DB_OK;
		}
	}

    // }}}
    // {{{ autoCommit()

	/**
	 * Enable/disable automatic commits [not supported by Virt]
	 *
	 * @access public
	 *
	 * @return object DB error code (not capable)
	 */
	function autoCommit($onoff = false) {
		return $this->raiseError(DB_ERROR_NOT_CAPABLE);
	}

    // }}}
    // {{{ commit()

	/**
	 * Commit transactions on the current connection [not supported by Virt]
	 *
	 * @access public
	 *
	 * @return object DB error code (not capable)
	 */
	function commit() {
		return $this->raiseError(DB_ERROR_NOT_CAPABLE);
	}

    // }}}
    // {{{ rollback()

	/**
	 * Roll back all uncommitted transactions on the current connection.
	 * [not supported by Virt]
	 *
	 * @access public
	 *
	 * @return object DB error code (not capable)
	 */
	function rollback() {
		return $this->raiseError(DB_ERROR_NOT_CAPABLE);
	}

    // }}}

	// TODO/wishlist:
	// simpleFetch
	// simpleGet
	// affectedRows
    // longReadlen
    // binmode

	// Overwrite the __sleep() function in DB_common to avoid serialization problem
	function __sleep() {
		return array_keys(get_object_vars(&$this));
	}
}

// Local variables:
// tab-width: 4
// c-basic-offset: 4
// End:
?>
