<?php include("bee/compiler.php"); return; ?>
<script language="bee">

/************************************************************
 *	mailobj: Mail Object
 *
 *	arg%to = the receipient email address(es)
 *	arg%subject = the subject
 *	arg%from = the sender email address
 *	arg%fromname = the sender name
 *	arg%cc = the CC receipient email address(es)
 *	arg%bcc = the BCC receipient email address(es)
 *	arg%errorsto = the email address that errors will send to
 *	arg%ignore = fields to ignore (if body is null)
 *	arg%header = the extra header
 *	arg%mimetype = Content-Type - ignored if multipart is defined
 *	arg%encoding = Content-Transfer-Encoding
 *	arg%multipart = Will cause Content-Type: multipart/...
 *		The value can be one of mixed, alternative, related, digest, parallel.  Default is mixed.
 *	arg%body = the message body
 *
 *
 *	Methods
 *
 *	%addpart
 *		content="..."
 *		mimetype=".../..."
 *		encoding="..."
 *		charset="..."
 *		name="..."
 *		filename="..."
 *		description="..."
 *		cid="..."
 *		header="...";
 *		// charset is for "text/..." mimetype
 *		// filename will cause Content-Disposition: attachment; filename="..."
 *		// description will cause Content-Description: ...
 *		// cid will cause Content-ID: <...>
 *	%send;
 *		// Send the email
 *
 *
 *	Internal Variables
 *
 *	%part - part array
 *	%boundary - The boundary string
 *	%encodedlinelength - The encoded line length
 *
 *	If there are no parts and no "body", the sys%form will be taken.
 *		This is in conformance with the mailto interface.
 *	If there are parts, "body" is still used and default to "This is a multi-part message in MIME format.".
 *	However, "body" will be ignored by the email reader anyway, whether it is the default or not.
 *
 *
 *
 *	Example:
 *
 *	mo% = new mailobj;
 *	mo%to = "some@random.com";
 *	mo%subject = "Mail Object Example";
 *	mo%from = "me@mydomain.com";
 *	mo%fromname = "John Chen";
 *	mo%cc = "everybody@theworld.com, nobody@anywhere.com";
 *	mo%bcc = "big@brother.com";
 *	mo%multipart = "mixed";  // Parts presented in serial
 *
 *	mo%addpart
 *		content="{file%my.doc&content}"
 *		mimetype="application/msword"
 *		encoding="base64"
 *		name="my.doc"
 *		filename="my.doc";
 *
 *	mo%addpart
 *		content="{file%photo.jpg&content}"
 *		mimetype="image/jpeg"
 *		encoding="base64"
 *		name="itisme.jpg"
 *		filename="photo.jpg";
 *
 *	mo%send;
 *
 ************************************************************/


function mailobj
{
	this%function:addpart = "mailobj_addpart";
	this%function:send = "mailobj_send";
	this%function:newboundary = "mailobj_newboundary";

	// Parameters can be reloaded
	if ({#this%to:} == 0) this%to = "{arg%to}";
	if ({#this%to:} == 0) this%to = "{arg%this}";
	if ({#this%subject:} == 0) this%subject = "{arg%subject}";
	if ({#this%from:} == 0) this%from = "{arg%from}";
	if ({#this%fromname:} == 0) this%fromname = "{arg%fromname}";
	if ({#this%cc:} == 0) this%cc = "{arg%cc}";
	if ({#this%bcc:} == 0) this%bcc = "{arg%bcc}";
	if ({#this%errorsto:} == 0) this%errorsto = "{arg%errorsto}";
	if ({#this%ignore:} == 0) this%ignore = "{arg%ignore}";
	if ({#this%header:} == 0) this%header = "{arg%header}";
	if ({#this%mimetype:} == 0) this%mimetype = "{arg%mimetype}";
	if ({#this%encoding:} == 0) this%encoding = "{arg%encoding}";
	if ({#this%multipart:} == 0) this%multipart = "{arg%multipart}";
	if ({#this%body:} == 0) this%body = "{arg%body}";

	this%status = 0;
	this%message = "";

	this%part = "(array)";
	this%numboundaries = 0;

	this%encodedlinelength = 72;
	this%newboundary;
}


/************************************************************
 *	mailobj_addpart: Contruct a "body part"
 *
 *	%addpart content="..." mimetype=".../..." encoding="..." name="..." filename="..." description="..." cid="..." header="...";
 *
 ************************************************************/

function mailobj_addpart
{
	// At least one \n before AND after thisPart must be guaranteed after each argument processing
	thisPart = "\n";
	thisPart = "{thisPart}--{this%boundary}\n";

	myName = "{arg%name}";

	myFilename = "{arg%filename}";
	if ({#myFilename:} == 0) myFilename = "{myName}";

	myType = "{arg%mimetype|default:text/plain}";
	thisPart = "{thisPart}Content-Type: {myType}";
	if ({myType|foundre:^text/}) {
		if ({#arg%charset:}) thisPart = "{thisPart};\n\tcharset=\"{arg%charset}\"";
	} else {
		if ({#myName:}) thisPart = "{thisPart};\n\tname=\"{myName}\"";
	}
	thisPart = "{thisPart}\n";

	if ({#arg%encoding:}) thisPart = "{thisPart}Content-Transfer-Encoding: {arg%encoding}\n";
	if ({#arg%description:}) thisPart = "{thisPart}Content-Description: {arg%description}\n";
	if ({#arg%cid:}) thisPart = "{thisPart}Content-ID: <{arg%cid}>\n";
	if ({#myFilename:}) thisPart = "{thisPart}Content-Disposition: attachment;\n\tfilename=\"{myFilename}\"\n";

	if ({#arg%header:}) thisPart = "{thisPart}{arg%header}\n";

	// Eliminate blank lines and guarantee double \n at the end
	thisPart = "{thisPart|xreplacere:\n+,\n}";
	if ('{arg%encoding}' == 'base64') {
		encStr = "{arg%content|base64encode}";
		ell = "{this%encodedlinelength}";
		if ({ell}) {
			// Chop it up to fixed length lines
			encStrOrig = "{encStr}";
			encStr = "";
			for (i from 0 to {={#encStrOrig:}|dec} step {ell}) {
				encStr = "{encStr}{encStrOrig:|substr:{i},{ell}}\n";
			}
		}
		thisPart = "{thisPart}\n{encStr}";
	}
	else thisPart = "{thisPart}\n{arg%content}";

	this%part = (var)this%part conv="push:{thisPart}";
}



/************************************************************
 *	mailobj_send: Send the email
 *
 *	%send;
 *
 ************************************************************/

function mailobj_send
{
	// Must not change any internal variable used to build the message
	// to allow repeat sending from the same mailobj object.
	msg = "{this%body}";
	if ({#this%part} == 0) {
		hdr = "{this%header}";
		if ({#this%mimetype:}) hdr = "{hdr}\nContent-Type: {this%mimetype}\n";
	} else {
		if ({#msg:} == 0) msg = "This is a multi-part message in MIME format.\n";
		mtp = "mixed";
		switch ('{this%multipart|lowercase}') {
		case 'mixed':
		case 'alternative':
		case 'related':
		case 'digest':
		case 'parallel':
			mtp = "{this%multipart|lowercase}";
			break;
		}
		hdr = "{this%header}\nContent-Type: multipart/{mtp};\n\tboundary=\"{this%boundary}\"\n";
		if ({#this%encoding:}) hdr = "{hdr}Content-Transfer-Encoding: {this%encoding}\n";

		foreach (this%part) msg = "{msg}{foreach}\n";
		msg = "{msg}\n--{this%boundary}\n";
	}

	// Header clean up
	hdr = "{hdr|xreplacere:\n+,\n}";
	hdr = "{hdr|xreplacere:^\n+}";
	hdr = "{hdr|xreplacere:\n+$}";

	mailto "{this%to}"
		subject="{this%subject}"
		from="{this%from}"
		fromname="{this%fromname}"
		cc="{this%cc}"
		bcc="{this%bcc}"
		errorsto="{this%errorsto}"
		ignore="{this%ignore}"
		header="{hdr}"
		body="{msg}";

	result%function = (var)result%mailto;

/*
display "<pre>\n";
display "to=[{this%to}]\n";
display "subject=[{this%subject}]\n";
display "from=[{this%from}]\n";
display "fromname=[{this%fromname}]\n";
display "cc=[{this%cc}]\n";
display "bcc=[{this%bcc}]\n";
display "errorsto=[{this%errorsto}]\n";
display "ignore=[{this%ignore}]\n";
display "header=[{hdr}]\n";
display "body=[{msg}]\n";
foreach (this%part) {
	display "part:{foreach:key}=[{foreach}]\n";
}
display "result%function=[{result%function|list}]\n";
display "</pre>\n";
 */
}


/************************************************************
 *	mailobj_newboundary: Regenerate a new boundary string and return it
 *
 ************************************************************/

function mailobj_newboundary
{
	this%boundary = "--------=BEE_MAILOBJ_10_{this%numboundaries}.{sys%time|crypt:MD5}.{sys%time}--";
	var this%numboundaries conv="inc";
	result%function = "{this%boundary}";
}

</script>
